# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


r"""Minimal Flask application example for development.

SPHINX-START

Run example development server:

.. code-block:: console

   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ ./app-fixtures.sh
   $ FLASK_APP=app.py flask run -p 5000

Open settings page to generate a token:

.. code-block:: console

   $ open http://127.0.0.1:5000/account/settings/applications

Login with:

    | username: admin\@inveniosoftware.org
    | password: 123456

Click on "New token" and compile the form:
insert the name "foobar", check scope "test:scope" and click "create".
The server will show you the generated Access Token.

Make a request to test the token:

.. code-block:: console

    export TOKEN=<generated Access Token>
    curl -i -X GET -H "Content-Type:application/json" http://127.0.0.1:5000/ \
        -H "Authorization:Bearer $TOKEN"

To end and remove any traces of example application, stop the example
application and run:
.. code-block:: console

   $ ./app-teardown.sh


SPHINX-END
"""

from __future__ import absolute_import, print_function

import os

from flask import Flask, render_template
from flask_breadcrumbs import Breadcrumbs
from flask_oauthlib.provider import OAuth2Provider
from invenio_access import InvenioAccess
from invenio_accounts import InvenioAccountsREST, InvenioAccountsUI
from invenio_accounts.views import blueprint as blueprint_account
from invenio_admin import InvenioAdmin
from invenio_admin.views import blueprint as blueprint_admin_ui
from invenio_assets import InvenioAssets
from invenio_db import InvenioDB
from invenio_i18n import InvenioI18N
from invenio_theme import InvenioTheme

from invenio_oauth2server import InvenioOAuth2Server, \
    InvenioOAuth2ServerREST, current_oauth2server, require_api_auth, \
    require_oauth_scopes
from invenio_oauth2server.models import Scope
from invenio_oauth2server.views import server_blueprint, settings_blueprint

# Create Flask application
app = Flask(__name__)
app.config.update(
    APP_ENABLE_SECURE_HEADERS=False,
    CELERY_ALWAYS_EAGER=True,
    CELERY_CACHE_BACKEND='memory',
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_RESULT_BACKEND='cache',
    OAUTH2_CACHE_TYPE='simple',
    OAUTHLIB_INSECURE_TRANSPORT=True,
    TESTING=True,
    DEBUG=True,  # needed to register localhost addresses as callback_urls
    SECRET_KEY='test_key',
    SECURITY_PASSWORD_HASH='plaintext',
    SECURITY_PASSWORD_SCHEMES=['plaintext'],
    SECURITY_DEPRECATED_PASSWORD_SCHEMES=[],
    LOGIN_DISABLED=False,
    TEMPLATE_AUTO_RELOAD=True,
    SQLALCHEMY_TRACK_MODIFICATIONS=True,
    SQLALCHEMY_DATABASE_URI=os.getenv(
        'SQLALCHEMY_DATABASE_URI', 'sqlite:///example.db'),
    I18N_LANGUAGES=[
        ('de', 'German'),
        ('es', 'Spanish'),
        ('fr', 'French'),
        ('it', 'Italian'),
    ],
)
InvenioAssets(app)
InvenioTheme(app)
InvenioI18N(app)
Breadcrumbs(app)
InvenioDB(app)
InvenioAdmin(app)
InvenioAccess(app)
OAuth2Provider(app)
InvenioOAuth2ServerREST(app)

accounts = InvenioAccountsUI(app)
InvenioAccountsREST(app)
app.register_blueprint(blueprint_account)

InvenioOAuth2Server(app)

# Register blueprints
app.register_blueprint(settings_blueprint)
app.register_blueprint(server_blueprint)
app.register_blueprint(blueprint_admin_ui)

with app.app_context():
    # Register a test scope
    current_oauth2server.register_scope(
        Scope('test:scope', help_text='Access to the homepage',
              group='test'))


# @app.route('/jwt', methods=['GET'])
# def jwt():
#     """JWT."""
#     return render_template('jwt.html')


@app.route('/', methods=['GET', 'POST'])
@require_api_auth()
@require_oauth_scopes('test:scope')
def index():
    """Protected endpoint."""
    return 'hello world'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that implements OAuth 2 server."""

from __future__ import absolute_import, print_function

import os
import warnings

import oauthlib.common as oauthlib_commmon
import pkg_resources
import six
from flask import abort, current_app, request, session
from flask_kvsession import KVSessionInterface
from flask_login import current_user
from flask_oauthlib.contrib.oauth2 import bind_cache_grant
from werkzeug.utils import cached_property, import_string

from . import config
from .models import OAuthUserProxy, Scope
from .provider import oauth2


class _OAuth2ServerState(object):
    """OAuth2 server state storing registered scopes."""

    def __init__(self, app, entry_point_group=None):
        """Initialize state."""
        self.app = app
        self.scopes = {}

        # Initialize OAuth2 provider
        oauth2.init_app(app)

        # Flask-OAuthlib does not support CACHE_REDIS_URL
        if app.config['OAUTH2_CACHE_TYPE'] == 'redis' and app.config.get(
                'CACHE_REDIS_URL'):
            from redis import from_url as redis_from_url
            app.config.setdefault(
                'OAUTH2_CACHE_REDIS_HOST',
                redis_from_url(app.config['CACHE_REDIS_URL'])
            )

        # Configures an OAuth2Provider instance to use configured caching
        # system to get and set the grant token.
        bind_cache_grant(app, oauth2, lambda: OAuthUserProxy(current_user))

        # Disables oauthlib's secure transport detection in in debug mode.
        if app.debug or app.testing:
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

        if entry_point_group:
            self.load_entry_point_group(entry_point_group)

    def scope_choices(self, exclude_internal=True):
        """Return list of scope choices.

        :param exclude_internal: Exclude internal scopes or not.
            (Default: ``True``)
        :returns: A list of tuples (id, scope).
        """
        return [
            (k, scope) for k, scope in sorted(self.scopes.items())
            if not exclude_internal or not scope.is_internal
        ]

    def register_scope(self, scope):
        """Register a scope.

        :param scope: A :class:`invenio_oauth2server.models.Scope` instance.
        """
        if not isinstance(scope, Scope):
            raise TypeError("Invalid scope type.")
        assert scope.id not in self.scopes
        self.scopes[scope.id] = scope

    def load_entry_point_group(self, entry_point_group):
        """Load actions from an entry point group.

        :param entry_point_group: The entrypoint group name to load plugins.
        """
        for ep in pkg_resources.iter_entry_points(group=entry_point_group):
            self.register_scope(ep.load())

    def load_obj_or_import_string(self, value):
        """Import string or return object.

        :params value: Import path or class object to instantiate.
        :params default: Default object to return if the import fails.
        :returns: The imported object.
        """
        imp = self.app.config.get(value)
        if isinstance(imp, six.string_types):
            return import_string(imp)
        elif imp:
            return imp

    @cached_property
    def jwt_verification_factory(self):
        """Load default JWT verification factory."""
        return self.load_obj_or_import_string(
            'OAUTH2SERVER_JWT_VERIFICATION_FACTORY'
        )


class InvenioOAuth2Server(object):
    """Invenio-OAuth2Server extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self._state = self.init_app(app, **kwargs)

    def init_app(self, app, entry_point_group='invenio_oauth2server.scopes',
                 **kwargs):
        """Flask application initialization.

        :param app: An instance of :class:`flask.Flask`.
        :param entry_point_group: The entrypoint group name to load plugins.
            (Default: ``'invenio_oauth2server.scopes'``)
        """
        self.init_config(app)

        state = _OAuth2ServerState(app, entry_point_group=entry_point_group)

        app.extensions['invenio-oauth2server'] = state
        return state

    def init_config(self, app):
        """Initialize configuration.

        :param app: An instance of :class:`flask.Flask`.
        """
        app.config.setdefault(
            'OAUTH2SERVER_BASE_TEMPLATE',
            app.config.get('BASE_TEMPLATE',
                           'invenio_oauth2server/base.html'))
        app.config.setdefault(
            'OAUTH2SERVER_COVER_TEMPLATE',
            app.config.get('COVER_TEMPLATE',
                           'invenio_oauth2server/base.html'))
        app.config.setdefault(
            'OAUTH2SERVER_SETTINGS_TEMPLATE',
            app.config.get('SETTINGS_TEMPLATE',
                           'invenio_oauth2server/settings/base.html'))

        for k in dir(config):
            if k.startswith('OAUTH2SERVER_') or k.startswith('OAUTH2_'):
                app.config.setdefault(k, getattr(config, k))

    def __getattr__(self, name):
        """Proxy to state object."""
        return getattr(self._state, name, None)


def verify_oauth_token_and_set_current_user():
    """Verify OAuth token and set current user on request stack.

    This function should be used **only** on REST application.

    .. code-block:: python

        app.before_request(verify_oauth_token_and_set_current_user)
    """
    for func in oauth2._before_request_funcs:
        func()

    if not hasattr(request, 'oauth') or not request.oauth:
        scopes = []
        try:
            valid, req = oauth2.verify_request(scopes)
        except ValueError:
            abort(400, 'Error trying to decode a non urlencoded string.')

        for func in oauth2._after_request_funcs:
            valid, req = func(valid, req)

        if valid:
            request.oauth = req


class InvenioOAuth2ServerREST(object):
    """Invenio-OAuth2Server REST extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        """Flask application initialization.

        :param app: An instance of :class:`flask.Flask`.
        """
        self.init_config(app)

        allowed_urlencode_chars = app.config.get(
            'OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS')
        if allowed_urlencode_chars:
            InvenioOAuth2ServerREST.monkeypatch_oauthlib_urlencode_chars(
                allowed_urlencode_chars)
        app.before_request(verify_oauth_token_and_set_current_user)

    def init_config(self, app):
        """Initialize configuration."""
        app.config.setdefault(
            'OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS',
            getattr(config, 'OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS'))

    @staticmethod
    def monkeypatch_oauthlib_urlencode_chars(chars):
        """Monkeypatch OAuthlib set of "URL encoded"-safe characters.

        .. note::

            OAuthlib keeps a set of characters that it considers as valid
            inside an URL-encoded query-string during parsing of requests. The
            issue is that this set of characters wasn't designed to be
            configurable since it should technically follow various RFC
            specifications about URIs, like for example `RFC3986
            <https://www.ietf.org/rfc/rfc3986.txt>`_. Many online services and
            frameworks though have designed their APIs in ways that aim at
            keeping things practical and readable to the API consumer, making
            use of special characters to mark or seperate query-string
            arguments. Such an example is the usage of embedded JSON strings
            inside query-string arguments, which of course have to contain the
            "colon" character (:) for key/value pair definitions.

            Users of the OAuthlib library, in order to integrate with these
            services and frameworks, end up either circumventing these "static"
            restrictions of OAuthlib by pre-processing query-strings, or -in
            search of a more permanent solution- directly make Pull Requests
            to OAuthlib to include additional characters in the set, and
            explain the logic behind their decision (one can witness these
            efforts inside the git history of the source file that includes
            this set of characters `here
            <https://github.com/idan/oauthlib/commits/master/oauthlib/common.py>`_).
            This kind of tactic leads easily to misconceptions about the
            ability one has over the usage of specific features of services and
            frameworks. In order to tackle this issue in Invenio-OAuth2Server,
            we are monkey-patching this set of characters using a configuration
            variable, so that usage of any special characters is a conscious
            decision of the package user.
        """
        modified_chars = set(chars)
        always_safe = set(oauthlib_commmon.always_safe)
        original_special_chars = oauthlib_commmon.urlencoded - always_safe
        if modified_chars != original_special_chars:
            warnings.warn(
                'You are overriding the default OAuthlib "URL encoded" set of '
                'valid characters. Make sure that the characters defined in '
                'oauthlib.common.urlencoded are indeed limitting your needs.',
                RuntimeWarning
            )
            oauthlib_commmon.urlencoded = always_safe | modified_chars



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuth2Server decorators test cases."""

from datetime import datetime

from flask import url_for
from invenio_accounts.proxies import current_accounts


def test_require_api_auth_oauthlib_urldecode_issue(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.get(app.url_for_test1resource, query_string='q=k:v')
        assert 401 == res.status_code


def test_require_api_auth_test1(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.get(app.url_for_test1resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers
        res = client.get(app.url_for_test1resource_token)
        assert 200 == res.status_code
        assert 'Set-Cookie' not in res.headers


def test_require_api_auth_test2(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.get(app.url_for_test2resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers
        res = client.get(app.url_for_test2resource_token)
        assert 200 == res.status_code
        assert 'Set-Cookie' not in res.headers


def test_require_oauth_scopes_test1(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.post(app.url_for_test1resource_token)
        assert 200 == res.status_code
        assert 'Set-Cookie' not in res.headers
        res = client.post(app.url_for_test1resource_token_noscope)
        assert 403 == res.status_code
        assert 'Set-Cookie' not in res.headers


def test_require_oauth_scopes_test2(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.post(app.url_for_test2resource_token)
        assert 200 == res.status_code
        assert 'Set-Cookie' not in res.headers
        res = client.post(app.url_for_test2resource_token_noscope)
        assert 403 == res.status_code
        assert 'Set-Cookie' not in res.headers


def test_require_oauth_scopes_allow_anonymous(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.get(app.url_for_test4resource)
        assert 200 == res.status_code
        assert b'None' == res.data
        assert 'Set-Cookie' not in res.headers

        res = client.get(app.url_for_test4resource_token)
        assert 200 == res.status_code
        assert u'{0}'.format(app.user_id).encode('utf-8') == res.data
        assert 'Set-Cookie' not in res.headers


def test_rest_extension(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        res = client.post(app.url_for_test4resource)
        assert 200 == res.status_code
        assert b'None' == res.data
        assert 'Set-Cookie' not in res.headers

        res = client.post(app.url_for_test4resource_token)
        assert 200 == res.status_code
        assert u'{0}'.format(app.user_id).encode('utf-8') == res.data
        assert 'Set-Cookie' not in res.headers


def test_access_login_required(resource_fixture):
    app = resource_fixture
    with app.test_client() as client:
        # try to access to authentication required zone
        res = client.post(app.url_for_test3resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers
        # try to access a scope protected zone (and pass)
        res = client.post(app.url_for_test2resource_token)
        assert 200 == res.status_code
        # try to access to authentication required zone
        res = client.post(app.url_for_test3resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers
        # try to access a scope protected zone (and fail)
        res = client.post(app.url_for_test2resource_token_noscope)
        assert 403 == res.status_code
        assert 'Set-Cookie' not in res.headers
        # try to access to login_required zone (and redirected to login)
        res = client.post(app.url_for_test3resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers
        # login
        res = client.post(url_for('security.login'), data=dict(
            email='info@inveniosoftware.org',
            password='tester'
        ))
        assert 'Set-Cookie' in res.headers
        # logout
        res = client.get(url_for('security.logout'))
        assert 302 == res.status_code

        res = client.post(app.url_for_test2resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers

        # try to access to login_required zone (and pass)
        res = client.post(app.url_for_test2resource_token)
        assert 200 == res.status_code
        assert 'Set-Cookie' not in res.headers
        # try to access to login_required zone (and not pass)
        res = client.post(app.url_for_test3resource)
        assert 401 == res.status_code
        assert 'Set-Cookie' not in res.headers


def test_jwt_client(resource_fixture, api_app):
    """Test client."""
    app = resource_fixture
    # Enable JWT
    app.config['ACCOUNTS_JWT_ENABLE'] = True
    with app.test_client() as client:

        # Try to access to authentication required zone
        res = client.post(app.url_for_test3resource)
        assert 401 == res.status_code

        # Login
        res = client.post(url_for('security.login'), data=dict(
            email='info@inveniosoftware.org',
            password='tester'
        ))
        assert 'Set-Cookie' in res.headers
        # Try to access to without a JWT
        res = client.post(app.url_for_test3resource)
        assert 400 == res.status_code

        # Generate a token
        token = current_accounts.jwt_creation_factory()
        # Make the request
        res = client.post(
            app.url_for_test3resource,
            headers=[
                ('Authorization', 'Bearer {}'.format(token))
            ]
        )
        assert 200 == res.status_code

        # Try with invalid user
        token = current_accounts.jwt_creation_factory(user_id=-20)
        # Make the request
        res = client.post(
            app.url_for_test3resource,
            headers=[
                ('Authorization', 'Bearer {}'.format(token))
            ]
        )
        assert 403 == res.status_code
        assert 'The JWT token is not valid.' in res.get_data(as_text=True)

        # Try to access with expired token
        extra = dict(
            exp=datetime(1970, 1, 1),
        )
        # Create token
        token = current_accounts.jwt_creation_factory(additional_data=extra)
        # Make the request
        res = client.post(
            app.url_for_test3resource,
            headers=[
                ('Authorization', 'Bearer {0}'.format(token))
            ]
        )
        assert 'The JWT token is expired.' in res.get_data(as_text=True)

        # Not correct Schema
        # Generate a token
        token = current_accounts.jwt_creation_factory()
        # Make the request
        res = client.post(
            app.url_for_test3resource,
            headers=[
                ('Authorization', 'Avengers {}'.format(token))
            ]
        )
        assert 400 == res.status_code
        assert 'Missing required header argument.' in res.get_data(
            as_text=True)

        # Check different header type
        api_app.config['OAUTH2SERVER_JWT_AUTH_HEADER'] = 'X-Travis-Mark-XLII'
        api_app.config['OAUTH2SERVER_JWT_AUTH_HEADER_TYPE'] = None
        # Create token
        token = current_accounts.jwt_creation_factory()
        # Make the request
        res = client.post(
            app.url_for_test3resource,
            headers=[
                ('X-Travis-Mark-XLII', '{0}'.format(token))
            ]
        )
        assert 200 == res.status_code



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin views for managing access to actions."""

from flask_admin.contrib.sqla import ModelView
from invenio_db import db

from .models import Client, Token


def _(x):
    """Identity."""
    return x


class ClientView(ModelView):
    """View for managing access to actions by users."""

    can_delete = True
    can_create = False
    can_view_details = True
    column_display_all_relations = True

    list_all = ('name', 'description', 'website', 'user_id', 'client_id', )

    column_list = \
        column_searchable_list = \
        column_sortable_list = \
        column_details_list = \
        list_all

    column_list = list_all
    column_default_sort = ('client_id', True)


class TokenView(ModelView):
    """View for managing access to actions by users with given role."""

    can_delete = True
    can_create = False
    can_view_details = True
    list_all = ('id', 'client_id', 'user_id', 'token_type', 'expires', )
    column_list = \
        column_searchable_list = \
        column_sortable_list = \
        column_details_list = \
        list_all


oauth2server_clients_adminview = {
    'view_class': ClientView,
    'args': [Client, db.session],
    'kwargs': {
        'name': _('OAuth Applications'),
        'category': _('User Management'),
    }
}

oauth2server_tokens_adminview = {
    'view_class': TokenView,
    'args': [Token, db.session],
    'kwargs': {
        'name': _('OAuth Application Tokens'),
        'category': _('User Management'),
    }
}

__all__ = (
    'oauth2server_clients_adminview',
    'oauth2server_tokens_adminview',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Errors raised by Invenio-OAuth2Server."""

from __future__ import absolute_import, print_function

import json

from werkzeug.exceptions import HTTPException


class OAuth2ServerError(Exception):
    """Base class for errors in oauth2server module."""


class ScopeDoesNotExists(OAuth2ServerError):
    """Scope is not registered it scopes registry."""

    def __init__(self, scope, *args, **kwargs):
        """Initialize exception by storing invalid scope."""
        super(ScopeDoesNotExists, self).__init__(*args, **kwargs)
        self.scope = scope


class JWTExtendedException(HTTPException):
    """Base exception for all JWT errors."""

    errors = None

    def __init__(self, errors=None, **kwargs):
        """Initialize JWTExtendedException."""
        super(JWTExtendedException, self).__init__(**kwargs)

        if errors is not None:
            self.errors = errors

    def get_errors(self):
        """Get errors.

        :returns: A list containing a dictionary representing the errors.
        """
        return [e.to_dict() for e in self.errors] if self.errors else None

    def get_body(self, environ=None):
        """Get the request body."""
        body = dict(
            status=self.code,
            message=self.description,
        )

        errors = self.get_errors()
        if self.errors:
            body['errors'] = errors

        return json.dumps(body)

    def get_headers(self, environ=None):
        """Get a list of headers."""
        return [('Content-Type', 'application/json')]


class JWTDecodeError(JWTExtendedException):
    """Exception raised when decoding is failed."""

    code = 400
    description = 'The JWT token has invalid format.'


class JWTInvalidIssuer(JWTExtendedException):
    """Exception raised when the user is not valid."""

    code = 403
    description = 'The JWT token is not valid.'


class JWTExpiredToken(JWTExtendedException):
    """Exception raised when JWT is expired."""

    code = 403
    description = 'The JWT token is expired.'


class JWTInvalidHeaderError(JWTExtendedException):
    """Exception raised when header argument is missing."""

    code = 400
    description = 'Missing required header argument.'


class JWTNoAuthorizationError(JWTExtendedException):
    """Exception raised when permission denied."""

    code = 400
    description = "The JWT token is not valid."



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create oauth2server tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = '12a88921ada2'
down_revision = 'aa546f2a8d2f'
branch_labels = ()
depends_on = '9848d0149abd'


def upgrade():
    """Upgrade database."""
    op.create_table(
        'oauth2server_client',
        sa.Column('name', sa.String(length=40), nullable=True),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column(
            'website',
            sqlalchemy_utils.types.url.URLType(),
            nullable=True),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('client_id', sa.String(length=255), nullable=False),
        sa.Column('client_secret', sa.String(length=255), nullable=False),
        sa.Column(
            'is_confidential',
            sa.Boolean(name='is_confidential'),
            nullable=True),
        sa.Column(
            'is_internal',
            sa.Boolean(name='is_internal'),
            nullable=True),
        sa.Column('_redirect_uris', sa.Text(), nullable=True),
        sa.Column('_default_scopes', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['user_id'], [u'accounts_user.id'], ),
        sa.PrimaryKeyConstraint('client_id')
    )
    op.create_index(
        op.f('ix_oauth2server_client_client_secret'),
        'oauth2server_client',
        ['client_secret'],
        unique=True
    )
    op.create_table(
        'oauth2server_token',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('client_id', sa.String(length=255), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=True),
        sa.Column('token_type', sa.String(length=255), nullable=True),
        sa.Column(
            'access_token',
            sqlalchemy_utils.EncryptedType(),
            nullable=True),
        sa.Column(
            'refresh_token',
            sqlalchemy_utils.EncryptedType(),
            nullable=True),
        sa.Column('expires', sa.DateTime(), nullable=True),
        sa.Column('_scopes', sa.Text(), nullable=True),
        sa.Column(
            'is_personal',
            sa.Boolean(name='is_personal'),
            nullable=True),
        sa.Column(
            'is_internal',
            sa.Boolean(name='is_internal'),
            nullable=True),
        sa.ForeignKeyConstraint(
            ['client_id'], [u'oauth2server_client.client_id'], ),
        sa.ForeignKeyConstraint(
            ['user_id'], [u'accounts_user.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_index(
        'ix_oauth2server_token_access_token',
        'oauth2server_token',
        ['access_token'],
        unique=True,
        mysql_length=255)
    op.create_index(
        'ix_oauth2server_token_refresh_token',
        'oauth2server_token',
        ['refresh_token'],
        unique=True,
        mysql_length=255)


def downgrade():
    """Downgrade database."""
    op.drop_index(
        'ix_oauth2server_token_refresh_token',
        table_name='oauth2server_token')
    op.drop_index(
        'ix_oauth2server_token_access_token',
        table_name='oauth2server_token')
    op.drop_table('oauth2server_token')
    op.drop_index(
        op.f('ix_oauth2server_client_client_secret'),
        table_name='oauth2server_client')
    op.drop_table('oauth2server_client')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test server views."""

from __future__ import absolute_import, print_function

from flask_principal import AnonymousIdentity


def test_user_identity_init(resource_fixture):
    """Test that user identity is loaded properly when a token is used."""
    app = resource_fixture
    with app.test_client() as client:
        # test without token (anonymous user)
        request_res = client.get(app.url_for_test0resource)
        assert request_res.status_code == 200
        assert isinstance(app.identity, AnonymousIdentity)

        # test with token
        request_res = client.get(app.url_for_test0resource_token)
        assert request_res.status_code == 200
        assert app.identity.user.id == app.user_id



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Useful decorators for checking authentication and scopes."""

import sys
from functools import wraps

from flask import abort, current_app, request, session
from flask_login import current_user
from six import reraise
from werkzeug.exceptions import Unauthorized

from .provider import oauth2
from .proxies import current_oauth2server


def require_api_auth(allow_anonymous=False):
    """Decorator to require API authentication using OAuth token.

    :param allow_anonymous: Allow access without OAuth token
        (default: ``False``).
    """
    def wrapper(f):
        """Wrap function with oauth require decorator."""
        f_oauth_required = oauth2.require_oauth()(f)

        @wraps(f)
        def decorated(*args, **kwargs):
            """Require OAuth 2.0 Authentication."""
            if not hasattr(current_user, 'login_via_oauth2'):
                if not current_user.is_authenticated:
                    if allow_anonymous:
                        return f(*args, **kwargs)
                    abort(401)
                if current_app.config['ACCOUNTS_JWT_ENABLE']:
                    # Verify the token
                    current_oauth2server.jwt_verification_factory(
                        request.headers)
                # fully logged in with normal session
                return f(*args, **kwargs)
            else:
                # otherwise, try oauth2
                return f_oauth_required(*args, **kwargs)
        return decorated
    return wrapper


def require_oauth_scopes(*scopes):
    r"""Decorator to require a list of OAuth scopes.

    Decorator must be preceded by a ``require_api_auth()`` decorator.
    Note, API key authentication is bypassing this check.

    :param \*scopes: List of scopes required.
    """
    required_scopes = set(scopes)

    def wrapper(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # Variable requests.oauth is only defined for oauth requests (see
            # require_api_auth() above).
            if hasattr(request, 'oauth') and request.oauth is not None:
                token_scopes = set(request.oauth.access_token.scopes)
                if not required_scopes.issubset(token_scopes):
                    abort(403)
            return f(*args, **kwargs)
        return decorated
    return wrapper



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test case for rebuilding access tokens."""

import sys

import pytest
from invenio_db import db

from invenio_oauth2server.models import Token
from invenio_oauth2server.utils import rebuild_access_tokens


def test_rebuilding_access_tokens(models_fixture):
    """Test rebuilding access tokens with random new SECRET_KEY."""
    app = models_fixture
    with app.app_context():
        old_secret_key = app.secret_key
        tokens_before = Token.query.order_by(Token.id).all()

        # Changing application SECRET_KEY
        app.secret_key = "NEW_SECRET_KEY"
        db.session.expunge_all()

        # Asserting the decoding error occurs with the stale SECRET_KEY
        if sys.version_info[0] < 3:  # python 2
            token = Token.query.filter(Token.id == tokens_before[0].id).one()
            assert token.access_token != tokens_before[0].access_token
        else:  # python 3
            with pytest.raises(ValueError):
                Token.query.filter(Token.id == tokens_before[0].id).one()

        db.session.expunge_all()
        rebuild_access_tokens(old_secret_key)
        tokens_after = Token.query.order_by(Token.id).all()

        for token_before, token_after in list(zip(tokens_before,
                                                  tokens_after)):
            assert token_before.access_token == token_after.access_token
            assert token_before.refresh_token == token_after.refresh_token



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Minimal OAuth2 consumer implementation to demonstrate OAuth2 protocol

SPHINX-START

This example OAuth2 consumer application is used to fetch an OAuth2 access
token from example application.

| For more information about OAuth2 protocol see
| https://invenio-oauthclient.readthedocs.io/en/latest/overview.html

.. note:: Before continuing make sure example application is running.

|

Open settings page of example app to register a new OAuth2 application:

.. code-block:: console

   $ open http://127.0.0.1:5000/account/settings/applications

Login using:

    | **username:** admin\@inveniosoftware.org
    | **password:** 123456

Click on "New application" and compile registration form with following data:
    | **Name:**           foobar-app
    | **Description:**    An example OAuth2 consumer application
    | **Website URL:**    \http://127.0.0.1:5100/
    | **Redirect URIs:**  \http://127.0.0.1:5100/authorized
    | **Client Type:**    Confidential


Click register and example application will generate and show you
a Client ID and Client Secret.

Open another terminal and move to examples-folder.

Export these values using following environment variables before starting
the example consumer or change values of corresponding keys
in *examples/consumer.py* to match.

.. code-block:: console

    $ export CONSUMER_CLIENT_ID=<generated_client_id>
    $ export CONSUMER_CLIENT_SECRET=<generated_client_secret>

|

**LOGOUT admin\@inveniosoftware.org from example application:**

.. code-block:: console

    $ open http://127.0.0.1:5000/logout

|

Run the example consumer

.. code-block:: console

    $ FLASK_APP=consumer.py flask run -p 5100

Start OAuth authorization flow and you will be redirected to example
application for authentication and to authorize example consumer to
access your account details on example application.

Login to example application with:

    | **username:** reader\@inveniosoftware.org
    | **password:** 123456

Review the authorization request presented to you and authorize
the example consumer.

You will be redirected back to example consumer where you can see details
of the authorization token that example application generated to
example consumer.

.. note::

    In case the authorization flow ends in an error, you can usually see
    the error in query-part of the URL.

|

Using example consumer's UI you can request a new access token from example
application either by using a refresh token or by completing
the authorization flow again.

To manage settings of OAuth2 consumer at invenio-oauth2server
settings page, login with the account that registered the consumer,
admin\@inveniosoftware.org.

To review and possibly revoke permissions of OAuth2 consumer that has
been authorized to access resources login with the account that authorized
the consumer, reader\@inveniosoftware.org.


|


This example consumer is inspired by example presented in
requests-oauthlib documentation
(http://requests-oauthlib.rtfd.io/en/latest/examples/real_world_example_with_refresh.html)
and is based on example application(s) of flask-oauthlib:
(https://github.com/lepture/flask-oauthlib/tree/master/example)
(https://github.com/lepture/flask-oauthlib/tree/master/example/contrib/experiment-client/douban.py)

Note that to support automatic refreshing of access tokens this consumer uses
flask-oauthlib.contrib.client which is considered experimental.

SPHINX-END
"""

from __future__ import print_function

import os
from functools import wraps
from pprint import pformat
from time import time

from flask import Flask, redirect, request, session, url_for
from flask_oauthlib.contrib.client import OAuth
from requests_oauthlib import OAuth2Session

# OAuth2 consumer configuration of this example consumer application
# You get the _CONSUMER_CLIENT_ID and _CONSUMER_CLIENT_SECRET when
# registering the consumer application to provider.
_CONSUMER_CLIENT_ID = 'insert_generated_client_id_here'
_CONSUMER_CLIENT_SECRET = 'insert_generated_client_secret_here'

CONSUMER_CREDENTIALS = dict(
    INVENIO_EXAMPLE_CONSUMER_CLIENT_ID=os.environ.get(
        'CONSUMER_CLIENT_ID', _CONSUMER_CLIENT_ID),
    INVENIO_EXAMPLE_CONSUMER_CLIENT_SECRET=os.environ.get(
        'CONSUMER_CLIENT_SECRET', _CONSUMER_CLIENT_SECRET),
    INVENIO_EXAMPLE_CONSUMER_SCOPE=[
        'test:scope',
        # 'email',
    ]
)


def create_app():
    # OAUTHLIB_INSECURE_TRANSPORT needed in order to use HTTP-endpoints
    os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = 'true'

    app = Flask(__name__)
    app.config.update(CONSUMER_CREDENTIALS)
    app.secret_key = 'development'

    oauth = OAuth(app)

    # OAuth2 Provider configuration of invenio-oauth2server example app.
    remote = oauth.remote_app(
        name='invenio_example_consumer',
        # client_id='',
        # client_secret='',
        # scope=['test:scope', 'email'],
        version='2',
        endpoint_url='http://127.0.0.1:5000/',
        access_token_url='http://127.0.0.1:5000/oauth/token',
        refresh_token_url='http://127.0.0.1:5000/oauth/token',
        authorization_url='http://127.0.0.1:5000/oauth/authorize'
    )

    def oauth_token_required(f):
        """ Decorator that checks if consumer already has access_token

        Only checks the availability, not validity of the token.
        """
        @wraps(f)
        def decorated_function(*args, **kwargs):
            if get_oauth_token() is None:
                # return redirect(url_for('.index'))
                return '<br>' \
                       '<h2>T_T You don\'t have a token yet</h2>' \
                       '<a href="{}"> Get one from here</a>' \
                       .format(url_for('index'))
            return f(*args, **kwargs)
        return decorated_function

    @app.before_first_request
    def _run_on_start():
        print("Client_ID is {}".format(remote.client_id))
        print("Client_Secret is {}".format(remote.client_secret))
        print("Redirect url is {}".format(url_for('authorized',
                                                  _external=True)))
        print("Requested scopes: {}".format(remote.scope))

    @app.route('/', methods=['GET'])
    def index():
        """ Mainpage. Display info about token and an action menu.
        """
        if get_oauth_token():
            return """
            <h1>^_^ Congratulations, you (already?) have an OAuth 2 token!</h1>
            <h2>What would you like to do next?</h2>
            <ul>
              <li><a href="/auto_refresh"> Implicitly refresh access token</a></li>
              <li><a href="/manual_refresh"> Explicitly refresh access token</a></li>
              <li><a href="/validate"> Validate the current token</a></li>
              <li><a href="/login"> Authorize the consumer again</a></li>
              <li><a href="/logout"> Delete current token, i.e Logout</a></li>
            </ul>
            <br> Current token information: <br> <pre>{}</pre>
            """.format(pformat(get_oauth_token(), indent=4))  # noqa

        return """
            <h1>OAuth2 consumer application</h1>
            <h2>Please click the link to start OAuth authorization flow</h2>
            <ul>
              <li><a href="/login"> Start authorization flow</a></li>
            </ul>
            """

    @app.route('/login', methods=['GET'])
    def login():
        """ Prepare an authorization request and redirect client to Provider

        Prepare an authorization request and redirect client / browser to
        Provider for authentication and to answer the authorization request.

        Example of key-values sent to Provider in query part URI
        - response_type: 'code' for authz flow, 'token' for implicit flow
        - client_id: 'identifier of consumer generated by provider'
        - redirect_uri: 'where client should be redirected after authorization'
        - scope: [list of scopes consumer requesting authorization for]
        - state: used to protect against CSRF

        """
        return remote.authorize(callback_uri=url_for('authorized',
                                                     _external=True))

    @app.route('/authorized', methods=['GET', 'POST'])
    def authorized():
        """Endpoint where client is redirected after an authorization attempt.

        OAuth2 Provider will redirect client / browser to this endpoint
        after an authorization attempt has been made.

        If authorization was granted redirect request contains
        'code'-parameter in querystring.
        Consumer automatically requests an access token using this
        'code'-parameter form Provider's access_token_url

        Example of key values sent in access token request:
        - code: 'value of code parameter obtained previously'
        - client_id: 'identifier of consumer generated by provider'
        - client_secret: 'secret of consumer generated by provider'
        - grant_type: 'authorization code' as code
        - redirect_url: used only to verify client is

        Example of response received from Provider
        - access_token: 2YotnFZFEjr1zCsicMWpAA
        - token_type: Bearer
        - expires_in: 3600
        - refresh_token: tGzv3JOkF0XG5Qx2TlKWIA
        - scopes: [list of scopes consumer is authorized to]

        Possible errors in authorization process are returned in querystring.

        """
        response = remote.authorized_response()
        if response is None:
            return '<br>' \
                   '<h2>T_T Access denied</h2>' \
                   '<br>' \
                   '<pre>error={error}</pre>'\
                   '<br><a href="{link}"> Try again</a>' \
                    .format(link=url_for('index'), error=request.args['error'])
        if isinstance(response, dict) and 'access_token' in response:
            store_oauth_token(response)
            return redirect(url_for('.index'))
        return str(response)

    @app.route("/auto_refresh", methods=['GET'])
    @oauth_token_required
    def auto_refresh():
        """Obtaining a new access token automatically using a refresh token.

        By making a call to this endpoint the consumer set token to expired
        state and upon making a new request for protected resource, consumer
        first request a new access token using refresh token.

        Note that all requests and updates related to tokens happen
        behind the scenes.
        """

        # We force expiration of the token by setting expired_at value
        # of the token to a past moment.
        # This will trigger an automatic refresh next time we interact with
        # Invenio API.
        token = get_oauth_token()
        token['expires_at'] = time() - 10

        resp = remote.get('oauth/info')  # prefixed with remote.endpoint_url
        return redirect(url_for('.index'))

    @app.route("/manual_refresh", methods=['GET'])
    @oauth_token_required
    def manual_refresh():
        """Obtaining a new access token manually using a refresh token.

        Request for new access token can be sent even if the old
        access token has not yet expired.

        Unfortunately flask-oauthlib OAuthApplication doesn't provide
        manual refreshing of tokens out-of-the-box, so internal
        OAuth2Session needs to be used.

        When using OAuth2Session to refresh, client_id and client_secret
        must be provided as extra arguments.
        When doing implicit / automatic refresh with flask-oauthlib
        these values are automatically included in the request.

        Note that obtained new access token must be stored manually.
        """

        extra = {
            'client_id': remote.client_id,
            'client_secret': remote.client_secret,
        }

        oauthsession = OAuth2Session(remote.client_id, token=get_oauth_token())

        resp = oauthsession.refresh_token(remote.refresh_token_url, **extra)

        store_oauth_token(resp)

        # Alternative way
        # # Build OAuth2Session
        # oauthsession = remote.make_oauth_session(token=get_oauth_token())
        #
        # # Use the built OAuth2Session to execute refresh token request
        # # even though authorization token is still valid.
        # store_oauth_token(oauthsession.refresh_token(
        #     remote.refresh_token_url))

        return redirect(url_for('.index'))

    @app.route("/validate", methods=['GET'])
    @oauth_token_required
    def validate():
        """ Endpoint used to check validity of access token.

        Basically consumer tries to access a resource in Provider and
        if the request succeeds the token is considered to be valid.
        """
        resp = remote.get('oauth/info')  # prefixed with remote.endpoint_url
        return """Endpoint '/oauth/info' returned following information:
            <pre>{}</pre>
            <br>
            <a href="/"> Return</a>
            """.format(pformat(resp.json(), indent=4))

    @app.route('/logout', methods=['GET'])
    def logout():
        """ Endpoint to logout from the consumer.

        Consumer will delete access and refresh tokens it has stored, losing
        access to resources at Provider.
        """

        session.pop('oauth_token', None)
        return redirect(url_for('index'))

    @remote.tokengetter
    def get_oauth_token():
        """ Returns the saved OAuth token object.

        Returned object contains all the information Provider sent
        regarding to access token.

        Example of OAuth token object:
          { "access_token": "6J3JyHLoP0K9HzL21V8y3Pj73zQDVf",
          "token_type": "Bearer",
          "expires_in": 3600,
          "refresh_token": "RxFZ1CSkjJF9PeFqDCaKTWYB6V5N4w",
          "scope": "test:scope" }
        """
        return session.get('oauth_token')

    @remote.tokensaver
    def store_oauth_token(resp):
        """ Saves the OAuth token and it's metadata obtained from Provider
        """
        # request-oauthlib expects a dictionary containing at least
        # access_token and token_type values. Successful access token response
        # must always contain those values, so we can store the whole response.
        session['oauth_token'] = resp

    return app

app = create_app()

if __name__ == '__main__':
    app = create_app()
    app.run(host='127.0.0.1', port=5100)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Validators for OAuth 2.0 redirect URIs and scopes."""

from __future__ import absolute_import, print_function

from flask import current_app
from oauthlib.oauth2.rfc6749.errors import InsecureTransportError, \
    InvalidRedirectURIError
from six.moves.urllib_parse import urlparse
from wtforms.validators import URL

from .errors import ScopeDoesNotExists
from .proxies import current_oauth2server


def validate_redirect_uri(value):
    """Validate a redirect URI.

    Redirect URIs must be a valid URL and use https unless the host is
    localhost for which http is accepted.

    :param value: The redirect URI.
    """
    sch, netloc, path, par, query, fra = urlparse(value)
    if not (sch and netloc):
        raise InvalidRedirectURIError()
    if sch != 'https':
        if ':' in netloc:
            netloc, port = netloc.split(':', 1)
        if not (netloc in ('localhost', '127.0.0.1') and sch == 'http'):
            raise InsecureTransportError()


def validate_scopes(value_list):
    """Validate if each element in a list is a registered scope.

    :param value_list: The list of scopes.
    :raises invenio_oauth2server.errors.ScopeDoesNotExists: The exception is
        raised if a scope is not registered.
    :returns: ``True`` if it's successfully validated.
    """
    for value in value_list:
        if value not in current_oauth2server.scopes:
            raise ScopeDoesNotExists(value)
    return True


class URLValidator(URL):
    """URL validator."""

    def __call__(self, form, field):
        """Check URL."""
        parsed = urlparse(field.data)
        if current_app.debug and parsed.hostname == 'localhost':
            return
        super(URLValidator, self).__call__(form=form, field=field)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuth2Server configuration variables."""

from datetime import timedelta

OAUTH2_CACHE_TYPE = 'redis'
"""Type of cache to use for storing the temporary grant token."""

OAUTH2_PROVIDER_ERROR_ENDPOINT = 'invenio_oauth2server.errors'
"""Error view endpoint."""

OAUTH2SERVER_CLIENT_ID_SALT_LEN = 40
"""Length of client id."""

OAUTH2SERVER_CLIENT_SECRET_SALT_LEN = 60
"""Length of the client secret."""

OAUTH2SERVER_TOKEN_PERSONAL_SALT_LEN = 60
"""Length of the personal access token."""

OAUTH2SERVER_ALLOWED_GRANT_TYPES = {
    'authorization_code', 'client_credentials', 'refresh_token',
}
"""A set of allowed grant types.

The allowed values are ``authorization_code``, ``password``,
``client_credentials``, ``refresh_token``). By default password is disabled,
as it requires the client application to gain access to the username and
password of the resource owner.
"""

OAUTH2SERVER_ALLOWED_RESPONSE_TYPES = {
    'code', 'token',
}
"""A set of allowed response types.

The allowed values are ``code`` and ``token``.

- ``code`` is used for authorization_code grant types
- ``token`` is used for implicit grant types
"""

OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS = '=&;:%+~,*@!()/?'
"""A string of special characters that should be valid inside a query string.

.. seealso::

    See :py:func:`monkeypatch_oauthlib_urlencode_chars
    <invenio_oauth2server.ext.InvenioOAuth2ServerREST.monkeypatch_oauthlib_urlencode_chars>`
    for a full explanation.
"""

OAUTH2SERVER_JWT_AUTH_HEADER = 'Authorization'
"""Header for the JWT.

.. note::

    Authorization: Bearer xxx
"""

OAUTH2SERVER_JWT_AUTH_HEADER_TYPE = 'Bearer'
"""Header Authorization type.

.. note::

    By default the authorization type is ``Bearer`` as recommented by
    `JWT  <https://jwt.io>`_
"""

OAUTH2SERVER_JWT_VERIFICATION_FACTORY = 'invenio_oauth2server.utils:' \
    'jwt_verify_token'
"""Import path of factory used to verify JWT.

The ``request.headers`` should be passed as parameter.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration of Flask-OAuthlib provider."""

from datetime import datetime, timedelta

from flask import _request_ctx_stack, current_app
from flask_login import current_user
from flask_oauthlib.provider import OAuth2Provider
from flask_principal import Identity, identity_changed
from flask_security.utils import verify_password
from invenio_db import db
from werkzeug.local import LocalProxy

from .models import Client, Token
from .scopes import email_scope

oauth2 = OAuth2Provider()
datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


@oauth2.usergetter
def get_user(email, password, *args, **kwargs):
    """Get user for grant type password.

    Needed for grant type 'password'. Note, grant type password is by default
    disabled.

    :param email: User email.
    :param password: Password.
    :returns: The user instance or ``None``.
    """
    user = datastore.find_user(email=email)
    if user and user.active and verify_password(password, user.password):
        return user


@oauth2.tokengetter
def get_token(access_token=None, refresh_token=None):
    """Load an access token.

    Add support for personal access tokens compared to flask-oauthlib.
    If the access token is ``None``, it looks for the refresh token.

    :param access_token: The access token. (Default: ``None``)
    :param refresh_token: The refresh token. (Default: ``None``)
    :returns: The token instance or ``None``.
    """
    if access_token:
        t = Token.query.filter_by(access_token=access_token).first()
    elif refresh_token:
        t = Token.query.join(Token.client).filter(
            Token.refresh_token == refresh_token,
            Token.is_personal == False,  # noqa
            Client.is_confidential == True,
        ).first()
    else:
        return None
    return t if t and t.user.active else None


@oauth2.clientgetter
def get_client(client_id):
    """Load the client.

    Needed for grant_type client_credentials.

    Add support for OAuth client_credentials access type, with user
    inactivation support.

    :param client_id: The client ID.
    :returns: The client instance or ``None``.
    """
    client = Client.query.get(client_id)
    if client and client.user.active:
        return client


@oauth2.tokensetter
def save_token(token, request, *args, **kwargs):
    """Token persistence.

    :param token: A dictionary with the token data.
    :param request: The request instance.
    :returns: A :class:`invenio_oauth2server.models.Token` instance.
    """
    # Exclude the personal access tokens which doesn't expire.
    user = request.user if request.user else current_user

    # Add user information in token endpoint response.
    # Currently, this is the only way to have the access to the user of the
    # token as well as the token response.
    token.update(user={'id': user.get_id()})

    # Add email if scope granted.
    if email_scope.id in token.scopes:
        token['user'].update(
            email=user.email,
            email_verified=user.confirmed_at is not None,
        )

    tokens = Token.query.filter_by(
        client_id=request.client.client_id,
        user_id=user.id,
        is_personal=False,
    )

    # make sure that every client has only one token connected to a user
    if tokens:
        for tk in tokens:
            db.session.delete(tk)
        db.session.commit()

    expires_in = token.get('expires_in')
    expires = datetime.utcnow() + timedelta(seconds=int(expires_in))

    tok = Token(
        access_token=token['access_token'],
        refresh_token=token.get('refresh_token'),
        token_type=token['token_type'],
        _scopes=token['scope'],
        expires=expires,
        client_id=request.client.client_id,
        user_id=user.id,
        is_personal=False,
    )
    db.session.add(tok)
    db.session.commit()
    return tok


@oauth2.after_request
def login_oauth2_user(valid, oauth):
    """Log in a user after having been verified."""
    if valid:
        oauth.user.login_via_oauth2 = True
        _request_ctx_stack.top.user = oauth.user
        identity_changed.send(current_app._get_current_object(),
                              identity=Identity(oauth.user.id))
    return valid, oauth



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility functions."""

from flask import current_app
from flask_login import current_user
from future.utils import raise_from
from invenio_accounts.errors import JWTDecodeError as _JWTDecodeError
from invenio_accounts.errors import JWTExpiredToken as _JWTExpiredToken
from invenio_accounts.utils import jwt_decode_token
from invenio_db.utils import rebuild_encrypted_properties

from .errors import JWTDecodeError, JWTExpiredToken, JWTInvalidHeaderError, \
    JWTInvalidIssuer
from .models import Token


def rebuild_access_tokens(old_key):
    """Rebuild the access_token field when the SECRET_KEY is changed.

    Needed to fix the access tokens used in the REST API calls.

    :param old_key: the old SECRET_KEY.
    """
    current_app.logger.info('rebuilding Token.access_token...')
    rebuild_encrypted_properties(old_key, Token,
                                 ['access_token', 'refresh_token'])


def jwt_verify_token(headers):
    """Verify the JWT token.

    :param dict headers: The request headers.
    :returns: The token data.
    :rtype: dict
    """
    # Get the token from headers
    token = headers.get(
        current_app.config['OAUTH2SERVER_JWT_AUTH_HEADER']
    )
    if token is None:
        raise JWTInvalidHeaderError
    # Get authentication type
    authentication_type = \
        current_app.config['OAUTH2SERVER_JWT_AUTH_HEADER_TYPE']
    # Check if the type should be checked
    if authentication_type is not None:
        # Get the prefix and the token
        prefix, token = token.split()
        # Check if the type matches
        if prefix != authentication_type:
            raise JWTInvalidHeaderError

    try:
        # Get the token data
        decode = jwt_decode_token(token)
        # Check the integrity of the user
        if current_user.get_id() != decode.get('sub'):
            raise JWTInvalidIssuer
        return decode
    except _JWTDecodeError as exc:
        raise_from(JWTDecodeError(), exc)
    except _JWTExpiredToken as exc:
        raise_from(JWTExpiredToken(), exc)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

from flask import Flask, url_for
from flask_admin import Admin
from invenio_db import db

from invenio_oauth2server import InvenioOAuth2Server
from invenio_oauth2server.admin import ClientView, TokenView, \
    oauth2server_clients_adminview, oauth2server_tokens_adminview


def test_admin(models_fixture):
    """Test flask-admin interface."""
    app = models_fixture
    InvenioOAuth2Server(app)

    assert isinstance(oauth2server_tokens_adminview, dict)
    assert isinstance(oauth2server_clients_adminview, dict)

    assert 'view_class' in oauth2server_tokens_adminview
    assert 'view_class' in oauth2server_clients_adminview

    admin = Admin(app, name="Test")

    clients_view = oauth2server_clients_adminview.pop('view_class')
    clients_model, clients_session = oauth2server_clients_adminview.pop('args')
    clients_kwargs = oauth2server_clients_adminview.pop('kwargs')
    tokens_view = oauth2server_tokens_adminview.pop('view_class')
    tokens_model, tokens_session = oauth2server_tokens_adminview.pop('args')
    tokens_kwargs = oauth2server_tokens_adminview.pop('kwargs')
    admin.add_view(clients_view(clients_model, db.session,
                                **clients_kwargs))
    admin.add_view(tokens_view(tokens_model, db.session,
                               **tokens_kwargs))

    menu_items = {str(item.name): item for item in admin.menu()}
    assert 'User Management' in menu_items
    assert menu_items['User Management'].is_category()

    submenu_items = {
        str(item.name): item for item in
        menu_items['User Management'].get_children()}
    assert 'OAuth Applications' in submenu_items
    assert 'OAuth Application Tokens' in submenu_items

    with app.test_request_context():
        token_request_url = url_for('token.index_view')
        client_request_url = url_for('client.index_view')
        client_view_url = url_for('invenio_oauth2server_settings.client_view',
                                  client_id='client_test_u1c1')
        client_reset_url = url_for(
            'invenio_oauth2server_settings.client_reset',
            client_id='client_test_u1c1')
        token_view_url = url_for(
            'invenio_oauth2server_settings.token_view',
            token_id='1')
        token_revoke_url = url_for(
            'invenio_oauth2server_settings.token_revoke',
            token_id='1')

    with app.app_context():
        with app.test_client() as client:
            res = client.get(
                token_request_url,
                follow_redirects=True
            )
            assert res.status_code == 200
            res = client.get(
                client_request_url,
                follow_redirects=True
            )
            assert res.status_code == 200
            res = client.get(
                client_view_url,
                follow_redirects=True
            )
            assert res.status_code == 200
            res = client.post(
                client_reset_url,
                follow_redirects=True
            )
            assert res.status_code == 200
            res = client.get(
                token_view_url,
                follow_redirects=True
            )
            assert res.status_code == 200
            res = client.post(
                token_revoke_url,
                follow_redirects=True
            )
            assert res.status_code == 405



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

from click.testing import CliRunner
from invenio_db import db

from invenio_oauth2server.cli import tokens_create, tokens_delete
from invenio_oauth2server.models import Client, Token


"""
$ invenio tokens create "my-token" \
    --email foobar@example.com --internal \
    --scopes user:email,records:read
abcdef1234567890abcdef1234567890

# Delete a token
$ invenio tokens delete --name "my-token" --email foobar@example.com
Token "abcdef1234567890abcdef1234567890" deleted

# ...or...
$ invenio tokens delete
Enter token: <password-like input>
Token "abcdef1234567890abcdef1234567890" deleted
"""


def test_cli_tokens(app, script_info, settings_fixture):
    """Test create user CLI."""
    runner = CliRunner()

    result = runner.invoke(
        tokens_create,
        ['--name', 'test-token',
         '--user', 'info@inveniosoftware.org'],
        obj=script_info)
    assert result.exit_code == 0
    access_token = result.output.strip()

    with app.app_context():
        client = Client.query.one()
        assert client.user.email == 'info@inveniosoftware.org'

        token = Token.query.one()
        assert token.access_token == access_token
        assert token.scopes == []

    result = runner.invoke(
        tokens_delete,
        ['--name', 'test-token',
         '--user', 'info@inveniosoftware.org',
         '--force'],
        obj=script_info)
    assert result.exit_code == 0

    with app.app_context():
        assert Token.query.count() == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test invenio_oauth2server validators."""

from __future__ import absolute_import, print_function

from collections import namedtuple

import pytest
from oauthlib.oauth2.rfc6749.errors import InsecureTransportError, \
    InvalidRedirectURIError
from wtforms.validators import ValidationError

from invenio_oauth2server.validators import URLValidator, validate_redirect_uri


@pytest.mark.parametrize('input,expected', [
    ('example.org/', InvalidRedirectURIError()),
    ('http://', InvalidRedirectURIError()),
    ('http://example.org/', InsecureTransportError()),
    ('https://example.org/', None),
    ('https://localhost/', None),
    ('https://127.0.0.1', None),
])
def test_validate_redirect_uri(input, expected):
    """Test redirect URI validator."""
    try:
        validate_redirect_uri(input)
    except Exception as e:
        assert type(e) is type(expected)


def test_url_validator(app):
    """Test url validator."""
    class Field(object):

        def __init__(self, data):
            self.data = data

        def gettext(self, *args, **kwargs):
            return 'text'

    with app.app_context():
        # usually if localhost, validation is failing
        with pytest.raises(ValidationError):
            URLValidator()(
                form=None, field=Field(data='http://localhost:5000'))
        URLValidator()(form=None, field=Field(data='http://mywebsite.it:5000'))

        # enable debug mode to accept also localhost
        app.config['DEBUG'] = True
        URLValidator()(form=None, field=Field(data='http://localhost:5000'))
        URLValidator()(form=None, field=Field(data='http://mywebsite.it:5000'))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that implements OAuth 2 server.

Protecting your REST API with authentication
--------------------------------------------

If you want to have your REST API endpoints protected using OAuth you
should register their blueprint inside the API app
(InvenioOAuth2ServerREST) which by default includes a ``before_request`` hook.
This hook will, if there is an OAuth token, verify it and set the current user
accordingly. It is important to highlight that this configuration allows
either authenticated clients or anonymous clients.

In case you need to allow access to a resource only for authenticated clients,
you should use the `require_api_auth
<api.html#invenio_oauth2server.decorators.require_api_auth>`__
decorator which requires OAuth2 login:

.. code-block:: python

    @app.route('/api/resource', methods=['GET'])
    @require_api_auth()
    def index():
        return 'Protected resource'

However, protecting your resources only with authentication is not
recommended. Instead, you should add an extra layer using always
`scopes <api.html#invenio_oauth2server.models.Scope>`__. This is because,
basically, any client that owns a token has control over every user resource.
Therefore, using scopes gives a fine-grain control. Here an example using
the default ``email_scope``:

.. code-block:: python

    from invenio_oauth2server.scopes import email_scope
    from flask_login import current_user

    @app.route('/api/email', methods=['GET'])
    @require_api_auth()
    @require_oauth_scopes(email_scope.id_)
    def index():
        return current_user.email

Delegating rights via scopes
----------------------------

As mentioned before, the recommended way to protect your endpoints is to use
fine-grain control with scopes. Invenio-OAuth2Server offers the possibility to
create new ones:

.. code-block:: python

    from invenio_oauth2server.models import Scope

    homepage_read = Scope('homepage:read',
                          help_text='Access to the homepage',
                          group='test')

Next, you should add them to ``setup.py`` entrypoints so they get initialized
at start up:

.. code-block:: python

    setup(
        ...
        entry_points={
            'invenio_oauth2server.scopes': [
                'homepage_read = path.to.scopes.file:homepage_read',
            ]
        }
        ...
    )

And then, they can be used in your application:

.. code-block:: python

    from path.to.scopes.file import homepage_read

    @app.route('/', methods=['GET'])
    @require_api_auth()
    @require_oauth_scopes(homepage_read.id_)
    def index():
        return 'Front page content.'

So, finally, with this example, we would allow any authenticated client with
rights to use the ``homepage_scope`` to read the homepage but, prevent from
reading the email if they do not have rights for using the ``email_scope``.

To test this features you can build your own application or use the provided
:doc:`example app </examplesapp>` as boilerplate.

Access control
--------------

It is important to remember that the usage of authentication and scopes is not
enough in most of the cases so access control need to be configured as well.
For more information about access control in Invenio you can visit
`Invenio-Access
<http://invenio-access.readthedocs.io/en/latest/>`__ documentation.


"""

from __future__ import absolute_import, print_function

from .decorators import require_api_auth, require_oauth_scopes
from .ext import InvenioOAuth2Server, InvenioOAuth2ServerREST
from .proxies import current_oauth2server
from .version import __version__

__all__ = (
    '__version__',
    'InvenioOAuth2Server',
    'InvenioOAuth2ServerREST',
    'require_api_auth',
    'require_oauth_scopes',
    'current_oauth2server',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprint instances."""

from __future__ import absolute_import, print_function

from .server import blueprint as server_blueprint
from .settings import blueprint as settings_blueprint

__all__ = (
    'server_blueprint',
    'settings_blueprint',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuth2 scopes."""

from __future__ import absolute_import, print_function

from flask_babelex import lazy_gettext as _

from invenio_oauth2server.models import Scope

email_scope = Scope(
    id_='user:email',
    group='user',
    help_text=_('Allow access to email address (read-only).'),
)
"""Scope to protect the user's email address."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add on delete cascade."""

from alembic import op

# revision identifiers, used by Alembic.
revision = '4e57407b8e4a'
down_revision = '12a88921ada2'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.drop_constraint('fk_oauth2server_client_user_id_accounts_user',
                       'oauth2server_client', type_='foreignkey')
    op.create_foreign_key(op.f('fk_oauth2server_client_user_id_accounts_user'),
                          'oauth2server_client', 'accounts_user', ['user_id'],
                          ['id'], ondelete='CASCADE')
    op.create_index(op.f('ix_oauth2server_client_user_id'),
                    'oauth2server_client', ['user_id'], unique=False)
    op.drop_constraint('fk_oauth2server_token_user_id_accounts_user',
                       'oauth2server_token', type_='foreignkey')
    op.create_foreign_key(op.f('fk_oauth2server_token_user_id_accounts_user'),
                          'oauth2server_token', 'accounts_user', ['user_id'],
                          ['id'], ondelete='CASCADE')
    op.create_index(op.f('ix_oauth2server_token_user_id'),
                    'oauth2server_token', ['user_id'], unique=False)
    op.drop_constraint('fk_oauth2server_token_client_id_oauth2server_client',
                       'oauth2server_token', type_='foreignkey')
    op.create_foreign_key(
        op.f('fk_oauth2server_token_client_id_oauth2server_client'),
        'oauth2server_token', 'oauth2server_client', ['client_id'],
        ['client_id'], ondelete='CASCADE')
    op.create_index(op.f('ix_oauth2server_token_client_id'),
                    'oauth2server_token', ['client_id'], unique=False)


def downgrade():
    """Downgrade database."""
    op.drop_constraint(op.f('fk_oauth2server_token_user_id_accounts_user'),
                       'oauth2server_token', type_='foreignkey')
    op.drop_index(op.f('ix_oauth2server_token_user_id'),
                  table_name='oauth2server_token')
    op.create_foreign_key('fk_oauth2server_token_user_id_accounts_user',
                          'oauth2server_token', 'accounts_user', ['user_id'],
                          ['id'])
    op.drop_constraint(
        op.f('fk_oauth2server_token_client_id_oauth2server_client'),
        'oauth2server_token', type_='foreignkey')
    op.drop_index(op.f('ix_oauth2server_token_client_id'),
                  table_name='oauth2server_token')
    op.create_foreign_key(
        'fk_oauth2server_token_client_id_oauth2server_client',
        'oauth2server_token', 'oauth2server_client', ['client_id'],
        ['client_id'])
    op.drop_constraint(op.f('fk_oauth2server_client_user_id_accounts_user'),
                       'oauth2server_client', type_='foreignkey')
    op.drop_index(op.f('ix_oauth2server_client_user_id'),
                  table_name='oauth2server_client')
    op.create_foreign_key('fk_oauth2server_client_user_id_accounts_user',
                          'oauth2server_client', 'accounts_user', ['user_id'],
                          ['id'])



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CLI commands for managing the OAuth server."""

from functools import wraps

import click
from flask.cli import with_appcontext
from invenio_accounts.models import User
from invenio_db import db
from werkzeug.local import LocalProxy

from .models import Client, Token
from .validators import validate_scopes


def lazy_result(f):
    """Decorate function to return LazyProxy."""
    @wraps(f)
    def decorated(ctx, param, value):
        return LocalProxy(lambda: f(ctx, param, value))
    return decorated


@lazy_result
def process_user(ctx, param, value):
    """Return a user if exists."""
    if value:
        if value.isdigit():
            user = User.query.get(str(value))
        else:
            user = User.query.filter(User.email == value).one_or_none()
        return user


@lazy_result
def process_scopes(ctx, param, values):
    """Return a user if exists."""
    validate_scopes(values)
    return values


@click.group()
def tokens():
    """OAuth2 server token commands."""


@tokens.command('create')
@click.option('-n', '--name', required=True)
@click.option(
    '-u', '--user', required=True, callback=process_user,
    help='User ID or email.')
@click.option(
    '-s', '--scope', 'scopes', multiple=True, callback=process_scopes)
@click.option('-i', '--internal', is_flag=True)
@with_appcontext
def tokens_create(name, user, scopes, internal):
    """Create a personal OAuth token."""
    token = Token.create_personal(
        name, user.id, scopes=scopes, is_internal=internal)
    db.session.commit()
    click.secho(token.access_token, fg='blue')


@tokens.command('delete')
@click.option('-n', '--name')
@click.option('-u', '--user', callback=process_user, help='User ID or email.')
@click.option('--token', 'read_access_token', is_flag=True)
@click.option('--force', is_flag=True)
@with_appcontext
def tokens_delete(name=None, user=None, read_access_token=None, force=False):
    """Delete a personal OAuth token."""
    if not (name or user) and not read_access_token:
        click.get_current_context().fail(
            'You have to pass either a "name" and "user" or the "token"')
    if name and user:
        client = Client.query.filter(
            Client.user_id == user.id,
            Client.name == name,
            Client.is_internal.is_(True)).one()
        token = Token.query.filter(
            Token.user_id == user.id,
            Token.is_personal.is_(True),
            Token.client_id == client.client_id).one()
    elif read_access_token:
        access_token = click.prompt('Token', hide_input=True)
        token = Token.query.filter(Token.access_token == access_token).one()
    else:
        click.get_current_context().fail('No token was found with provided')
    if force or click.confirm('Are you sure you want to delete the token?'):
        db.session.delete(token)
        db.session.commit()
        click.secho(
            'Token "{}" deleted.'.format(token.access_token), fg='yellow')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Define forms for generating access tokens and clients."""

from flask_babelex import lazy_gettext as _
from flask_wtf import FlaskForm as Form
from oauthlib.oauth2.rfc6749.errors import InsecureTransportError, \
    InvalidRedirectURIError
from wtforms import fields, validators, widgets
from wtforms.widgets import HTMLString
from wtforms_alchemy import model_form_factory

from .models import Client
from .validators import URLValidator, validate_redirect_uri


#
# Widget
#
def scopes_multi_checkbox(field, **kwargs):
    """Render multi checkbox widget."""
    kwargs.setdefault('type', 'checkbox')
    field_id = kwargs.pop('id', field.id)

    html = [u'<div class="row">']

    for value, label, checked in field.iter_choices():
        choice_id = u'%s-%s' % (field_id, value)

        options = dict(
            kwargs,
            name=field.name,
            value=value,
            id=choice_id,
            class_=' ',
        )

        if checked:
            options['checked'] = 'checked'

        html.append(u'<div class="col-md-3">')
        html.append(u'<label for="{0}" class="checkbox-inline">'.format(
            choice_id
        ))
        html.append(u'<input {0} /> '.format(widgets.html_params(**options)))
        html.append(u'{0} <br/><small class="text-muted">{1}</small>'.format(
            value, label.help_text
        ))
        html.append(u'</label></div>')
    html.append(u'</div>')

    return HTMLString(u''.join(html))


#
# Redirect URI field
#
class RedirectURIField(fields.TextAreaField):
    """Process redirect URI field data."""

    def process_formdata(self, valuelist):
        """Process form data."""
        if valuelist:
            self.data = '\n'.join([
                x.strip() for x in
                filter(lambda x: x, '\n'.join(valuelist).splitlines())
            ])

    def process_data(self, value):
        """Process data."""
        self.data = '\n'.join(value)


class RedirectURIValidator(object):
    """Validate if redirect URIs."""

    def __call__(self, form, field):
        """Call function."""
        errors = []
        for v in field.data.splitlines():
            try:
                validate_redirect_uri(v)
            except InsecureTransportError:
                errors.append(v)
            except InvalidRedirectURIError:
                errors.append(v)

        if errors:
            raise validators.ValidationError(
                _('Invalid redirect URIs: %(urls)s', urls=', '.join(errors))
            )


#
# Forms
#
class ClientFormBase(model_form_factory(Form)):
    """Base class for Client form."""

    class Meta:
        """Metadata class."""

        model = Client
        exclude = [
            'client_secret',
            'is_internal',
        ]
        strip_string_fields = True
        field_args = dict(
            website=dict(
                validators=[validators.DataRequired(), URLValidator()],
                widget=widgets.TextInput(),
            ),
        )


class ClientForm(ClientFormBase):
    """Client form."""

    # Trick to make redirect_uris render in the bottom of the form.
    redirect_uris = RedirectURIField(
        label=_('Redirect URIs (one per line)'),
        description=_(
            "One redirect URI per line. This is your application's"
            ' authorization callback URLs. HTTPS must be used for all '
            'hosts except localhost (for testing purposes).'),
        validators=[RedirectURIValidator(), validators.DataRequired()],
        default='',
    )

    is_confidential = fields.SelectField(
        label=_('Client type'),
        description=_(
            'Select confidential if your application is capable of keeping '
            'the issued client secret confidential (e.g. a web application), '
            'select public if your application cannot (e.g. a browser-based '
            'JavaScript application). If you select public, your application '
            'MUST validate the redirect URI.'),
        coerce=bool,
        choices=[(True, _('Confidential')), (False, _('Public'))],
        default=True,
    )


class TokenForm(Form):
    """Token form."""

    name = fields.StringField(
        description=_('Name of personal access token.'),
        validators=[
            validators.DataRequired(),
            validators.Length(
                max=40,
                message=_('The name must be less than 40 characters long.'))
        ]
    )

    scopes = fields.SelectMultipleField(
        widget=scopes_multi_checkbox,
        choices=[],  # Must be dynamically provided in view.
        description=_(
            'Scopes assign permissions to your personal access token.'
            ' A personal access token works just like a normal OAuth '
            ' access token for authentication against the API.')
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from __future__ import absolute_import, print_function

import pytest
from flask import Flask
from invenio_db import db

from invenio_oauth2server import InvenioOAuth2Server, InvenioOAuth2ServerREST
from invenio_oauth2server.ext import verify_oauth_token_and_set_current_user


def test_version():
    """Test version import."""
    from invenio_oauth2server import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = InvenioOAuth2Server(app)
    assert 'invenio-oauth2server' in app.extensions
    assert ext.app is app

    app = Flask('testapp')
    ext = InvenioOAuth2Server()
    assert 'invenio-oauth2server' not in app.extensions
    state = ext.init_app(app)
    assert 'invenio-oauth2server' in app.extensions
    assert state.app is app


def test_init_rest():
    """Test REST extension initialization."""
    app = Flask('testapp')
    ext = InvenioOAuth2ServerREST(app)
    assert verify_oauth_token_and_set_current_user in \
        app.before_request_funcs[None]

    app = Flask('testapp')
    ext = InvenioOAuth2ServerREST()
    assert verify_oauth_token_and_set_current_user not in \
        app.before_request_funcs.get(None, [])
    ext.init_app(app)
    assert verify_oauth_token_and_set_current_user in \
        app.before_request_funcs[None]


def test_init_rest_with_oauthlib_monkeypatch():
    """Test REST OAuthlib monkeypatching."""
    app = Flask('testapp')

    from oauthlib.common import urlencoded
    assert '^' not in urlencoded
    old_urlencoded = set(urlencoded)

    app.config['OAUTH2SERVER_ALLOWED_URLENCODE_CHARACTERS'] = '^'

    with pytest.warns(RuntimeWarning):
        InvenioOAuth2ServerREST(app)
    assert verify_oauth_token_and_set_current_user in \
        app.before_request_funcs[None]

    from oauthlib.common import urlencoded
    assert old_urlencoded != urlencoded
    assert '^' in urlencoded


def test_alembic(app):
    """Test alembic recipes."""
    ext = app.extensions['invenio-db']

    with app.app_context():
        if db.engine.name == 'sqlite':
            raise pytest.skip('Upgrades are not supported on SQLite.')

        assert not ext.alembic.compare_metadata()
        db.drop_all()
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()
        ext.alembic.downgrade(target='96e796392533')
        ext.alembic.upgrade()

        assert not ext.alembic.compare_metadata()
        ext.alembic.downgrade(target='96e796392533')
